@@ -1,6 +1,6 @@
 /* Threading support - implementation
  *
- * Copyright (C) 2018 - 2023, Stephan Mueller <smueller@chronox.de>
+ * Copyright (C) 2018 - 2024, Stephan Mueller <smueller@chronox.de>
  *
  * License: see LICENSE file in root directory
  *
@@ -30,7 +30,7 @@
 #include "atomic_bool.h"
 #include "bool.h"
 #include "config.h"
-#include "logger.h"
+#include "esdm_logger.h"
 #include "memset_secure.h"
 #include "mutex_w.h"
 #include "ret_checkers.h"
@@ -71,7 +71,6 @@ struct thread_ctx {
 					 * is ready for pickup? */
 
 	pthread_cond_t worker_cv;
-	pthread_mutex_t worker_lock;
 };
 
 /*
@@ -159,7 +158,6 @@ static inline void thread_cleanup_full(struct thread_ctx *tctx)
 	tctx->scheduled = false;
 	tctx->ret_ancestor = 0;
 	mutex_w_destroy(&tctx->inuse);
-	pthread_mutex_destroy(&tctx->worker_lock);
 }
 
 int thread_init(uint32_t groups)
@@ -169,9 +167,10 @@ int thread_init(uint32_t groups)
 	int ret;
 
 	if (groups > (THREADING_MAX_THREADS)) {
-		logger(LOGGER_ERR, LOGGER_C_THREADING,
-		       "Number of threads (%u) is less than the number of requested thread groups (%u)\n",
-		       THREADING_MAX_THREADS, groups);
+		esdm_logger(
+			LOGGER_ERR, LOGGER_C_THREADING,
+			"Number of threads (%u) is less than the number of requested thread groups (%u)\n",
+			THREADING_MAX_THREADS, groups);
 		return -EINVAL;
 	}
 
@@ -191,20 +190,19 @@ int thread_init(uint32_t groups)
 		atomic_bool_set_false(&threads[i].thread_pending);
 		mutex_w_init(&threads[i].inuse, false, 1);
 		atomic_bool_set_false(&threads[i].shutdown);
-		pthread_mutex_init(&threads[i].worker_lock, NULL);
 	}
 
 	threads_groups = groups;
 	threads_per_threadgroup = THREADING_MAX_THREADS / threads_groups;
 
-	logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
-	       "Initialized threading support for %u threads\n",
-	       THREADING_MAX_THREADS);
+	esdm_logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
+		    "Initialized threading support for %u threads\n",
+		    THREADING_MAX_THREADS);
 	if (threads_per_threadgroup * threads_groups < THREADING_MAX_THREADS) {
-		logger(LOGGER_WARN, LOGGER_C_THREADING,
-		       "%u thread slots will never be used\n",
-		       THREADING_MAX_THREADS -
-			       (threads_per_threadgroup * threads_groups));
+		esdm_logger(LOGGER_WARN, LOGGER_C_THREADING,
+			    "%u thread slots will never be used\n",
+			    THREADING_MAX_THREADS -
+				    (threads_per_threadgroup * threads_groups));
 	}
 
 out:
@@ -237,6 +235,7 @@ static void *thread_worker(void *arg)
 	while (1) {
 		mutex_w_lock(&tctx->inuse);
 
+locked:
 		if (atomic_bool_read(&tctx->shutdown)) {
 			/* Request for termination */
 			mutex_w_unlock(&tctx->inuse);
@@ -254,15 +253,16 @@ static void *thread_worker(void *arg)
 			/* Work to do, execute */
 			tctx->ret_ancestor = tctx->start_routine(tctx->data);
 			thread_cleanup(tctx);
-			logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
-			       "Thread %u completed\n", tctx->thread_num);
+			esdm_logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
+				    "Thread %u completed\n", tctx->thread_num);
 			mutex_w_unlock(&tctx->inuse);
 			pthread_cond_broadcast(&thread_wait_cv);
 		} else {
-			mutex_w_unlock(&tctx->inuse);
-
 			/* Idle */
-			thread_block(&tctx->worker_cv, &tctx->worker_lock);
+			/* inuse.lock is locked */
+			pthread_cond_wait(&tctx->worker_cv, &tctx->inuse.lock);
+			/* inuse.lock is locked */
+			goto locked;
 		}
 	}
 
@@ -329,9 +329,10 @@ static int thread_schedule(int (*start_routine)(void *), void *tdata,
 	unsigned int special_slot = thread_get_special_slot(thread_group);
 
 	if (threads_groups < thread_group && !special_slot) {
-		logger(LOGGER_ERR, LOGGER_C_THREADING,
-		       "undefined thread group requested (%u, max thread group is %u)\n",
-		       thread_group, threads_groups);
+		esdm_logger(
+			LOGGER_ERR, LOGGER_C_THREADING,
+			"undefined thread group requested (%u, max thread group is %u)\n",
+			thread_group, threads_groups);
 		return -EINVAL;
 	}
 
@@ -381,9 +382,10 @@ static int thread_schedule(int (*start_routine)(void *), void *tdata,
 				if (ret)
 					return ret;
 
-				logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
-				       "Thread %u for thread group %u allocated\n",
-				       i, thread_group);
+				esdm_logger(
+					LOGGER_VERBOSE, LOGGER_C_THREADING,
+					"Thread %u for thread group %u allocated\n",
+					i, thread_group);
 			}
 
 			/* Catch the return code of the ancestor thread */
@@ -394,9 +396,9 @@ static int thread_schedule(int (*start_routine)(void *), void *tdata,
 			 * Use the thread from the thread pool and schedule
 			 * job.
 			 */
-			logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
-			       "Thread %u for thread group %u assigned\n", i,
-			       thread_group);
+			esdm_logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
+				    "Thread %u for thread group %u assigned\n",
+				    i, thread_group);
 			threads[i].data = tdata;
 			threads[i].start_routine = start_routine;
 			threads[i].parent = pthread_self();
@@ -532,8 +534,8 @@ static int thread_wait_all(bool system_threads)
 			pthread_join(threads[i].thread_id, NULL);
 			ret |= threads[i].ret_ancestor;
 			thread_cleanup_full(&threads[i]);
-			logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
-			       "Thread %u terminated\n", i);
+			esdm_logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
+				    "Thread %u terminated\n", i);
 		}
 	}
 
@@ -568,8 +570,8 @@ static void thread_cancel(bool system_threads)
 			pthread_cancel(threads[i].thread_id);
 			pthread_join(threads[i].thread_id, NULL);
 			thread_cleanup_full(&threads[i]);
-			logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
-			       "Thread %u killed\n", i);
+			esdm_logger(LOGGER_VERBOSE, LOGGER_C_THREADING,
+				    "Thread %u killed\n", i);
 		}
 	}
 
