@@ -1,7 +1,7 @@
 /*
  * ESDM DRNG management
  *
- * Copyright (C) 2022 - 2023, Stephan Mueller <smueller@chronox.de>
+ * Copyright (C) 2022 - 2024, Stephan Mueller <smueller@chronox.de>
  *
  * License: see LICENSE file in root directory
  *
@@ -151,11 +151,11 @@ void esdm_drng_reset(struct esdm_drng *drng)
 	/* Ensure reseed during next call */
 	atomic_set(&drng->requests, 1);
 	atomic_set(&drng->requests_since_fully_seeded, 0);
-	drng->last_seeded = time(NULL);
+	clock_gettime(CLOCK_MONOTONIC, &drng->last_seeded);
 	drng->fully_seeded = false;
 	/* Do not set force, as this flag is used for the emergency reseeding */
 	drng->force_reseed = false;
-	logger(LOGGER_DEBUG, LOGGER_C_DRNG, "reset DRNG\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_DRNG, "reset DRNG\n");
 }
 
 /* Initialize the DRNG, except the mutex lock */
@@ -206,22 +206,24 @@ static int esdm_drng_mgr_selftest(void)
 	if (hash_cb->hash_selftest)
 		ret = hash_cb->hash_selftest();
 	else
-		logger(LOGGER_WARN, LOGGER_C_DRNG, "Hash self test missing\n");
+		esdm_logger(LOGGER_WARN, LOGGER_C_DRNG,
+			    "Hash self test missing\n");
 	mutex_reader_unlock(&drng->hash_lock);
 	CKINT_LOG(ret, "Hash self test failed: %d\n", ret);
-	logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-	       "Hash self test passed successfully\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_DRNG,
+		    "Hash self test passed successfully\n");
 
 	mutex_w_lock(&drng->lock);
 	drng_cb = drng->drng_cb;
 	if (drng_cb->drng_selftest)
 		ret = drng_cb->drng_selftest();
 	else
-		logger(LOGGER_WARN, LOGGER_C_DRNG, "DRNG self test missing\n");
+		esdm_logger(LOGGER_WARN, LOGGER_C_DRNG,
+			    "DRNG self test missing\n");
 	mutex_w_unlock(&drng->lock);
 	CKINT_LOG(ret, "DRNG self test failed: %d\n", ret);
-	logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-	       "DRNG self test passed successfully\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_DRNG,
+		    "DRNG self test passed successfully\n");
 
 out:
 	esdm_drng_put_instances();
@@ -256,29 +258,31 @@ int esdm_drng_mgr_initialize(void)
 	mutex_w_unlock(&esdm_drng_pr.lock);
 
 	if (!ret) {
-		logger(LOGGER_VERBOSE, LOGGER_C_DRNG,
-		       "DRNG with prediction resistance allocated\n");
+		esdm_logger(LOGGER_VERBOSE, LOGGER_C_DRNG,
+			    "DRNG with prediction resistance allocated\n");
 		mutex_w_init(&esdm_drng_init.lock, 1, 1);
 		ret = esdm_drng_alloc_common(&esdm_drng_init,
 					     esdm_default_drng_cb);
 		mutex_w_unlock(&esdm_drng_init.lock);
 		if (!ret) {
 			atomic_set(&esdm_avail, 2);
-			logger(LOGGER_VERBOSE, LOGGER_C_DRNG,
-			       "DRNG without prediction resistance allocated\n");
+			esdm_logger(
+				LOGGER_VERBOSE, LOGGER_C_DRNG,
+				"DRNG without prediction resistance allocated\n");
 		}
 	}
 
 	CKINT(ret);
 
-	logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-	       "ESDM for general use is available\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_DRNG,
+		    "ESDM for general use is available\n");
 
 	CKINT(esdm_drng_mgr_selftest());
 
 out:
-	if (ret)
+	if (ret) {
 		atomic_set(&esdm_avail, 0);
+	}
 	return ret;
 }
 
@@ -337,22 +341,28 @@ void esdm_set_reseed_max_time(uint32_t seconds)
 
 /************************* Random Number Generation ***************************/
 
-static bool esdm_time_after(time_t curr, time_t base)
+static bool esdm_time_after(struct timespec *curr, struct timespec *timeout)
 {
-	if (curr == (time_t)-1)
+	if (curr == NULL || timeout == NULL)
 		return false;
-	if (base == (time_t)-1)
+
+	if (curr->tv_sec > timeout->tv_sec)
 		return true;
-	return (curr > base) ? true : false;
+	if (curr->tv_sec == timeout->tv_sec && curr->tv_nsec > timeout->tv_nsec)
+		return true;
+
+	return false;
 }
 
-static time_t esdm_time_after_now(time_t base)
+static time_t esdm_time_after_now(struct timespec *timeout)
 {
-	time_t curr = time(NULL);
+	struct timespec curr;
 
-	if (curr == (time_t)-1)
+	if (clock_gettime(CLOCK_MONOTONIC, &curr) == -1)
 		return 0;
-	return esdm_time_after(curr, base) ? (curr - base) : 0;
+
+	return esdm_time_after(&curr, timeout) ?
+			       (curr.tv_sec - timeout->tv_sec) : 0;
 }
 
 /* Inject a data buffer into the DRNG - caller must hold its lock */
@@ -360,22 +370,23 @@ void esdm_drng_inject(struct esdm_drng *drng, const uint8_t *inbuf,
 		      size_t inbuflen, bool fully_seeded, const char *drng_type)
 {
 	BUILD_BUG_ON(ESDM_DRNG_RESEED_THRESH > INT_MAX);
-	logger(LOGGER_DEBUG, LOGGER_C_DRNG, "seeding %s DRNG with %zu bytes\n",
-	       drng_type, inbuflen);
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_DRNG,
+		    "seeding %s DRNG with %zu bytes\n", drng_type, inbuflen);
 
 	if (!drng->drng)
 		return;
 
 	if (drng->drng_cb->drng_seed(drng->drng, inbuf, inbuflen) < 0) {
-		logger(LOGGER_WARN, LOGGER_C_DRNG,
-		       "seeding of %s DRNG failed\n", drng_type);
+		esdm_logger(LOGGER_WARN, LOGGER_C_DRNG,
+			    "seeding of %s DRNG failed\n", drng_type);
 		drng->force_reseed = true;
 	} else {
 		int gc = ESDM_DRNG_RESEED_THRESH - atomic_read(&drng->requests);
 
-		logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-		       "%s DRNG stats since last seeding: %lu secs; generate calls: %d\n",
-		       drng_type, esdm_time_after_now(drng->last_seeded), gc);
+		esdm_logger(
+			LOGGER_DEBUG, LOGGER_C_DRNG,
+			"%s DRNG stats since last seeding: %lu secs; generate calls: %d\n",
+			drng_type, esdm_time_after_now(&drng->last_seeded), gc);
 
 		/* Count the numbers of generate ops since last fully seeded */
 		if (fully_seeded)
@@ -383,15 +394,16 @@ void esdm_drng_inject(struct esdm_drng *drng, const uint8_t *inbuf,
 		else
 			atomic_add(&drng->requests_since_fully_seeded, gc);
 
-		drng->last_seeded = time(NULL);
+		clock_gettime(CLOCK_MONOTONIC, &drng->last_seeded);
 		atomic_set(&drng->requests, ESDM_DRNG_RESEED_THRESH);
 		drng->force_reseed = false;
 
 		if (!drng->fully_seeded) {
 			drng->fully_seeded = fully_seeded;
 			if (drng->fully_seeded)
-				logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-				       "%s DRNG fully seeded\n", drng_type);
+				esdm_logger(LOGGER_DEBUG, LOGGER_C_DRNG,
+					    "%s DRNG fully seeded\n",
+					    drng_type);
 		}
 	}
 }
@@ -425,9 +437,10 @@ static uint32_t esdm_drng_seed_es_nolock(struct esdm_drng *drng, bool init_ops,
 		num_es_delivered = 0;
 
 		if (collected_entropy) {
-			logger(LOGGER_VERBOSE, LOGGER_C_DRNG,
-			       "Force fully seeding level for %s DRNG by repeatedly pull entropy from available entropy sources\n",
-			       drng_type);
+			esdm_logger(
+				LOGGER_VERBOSE, LOGGER_C_DRNG,
+				"Force fully seeding level for %s DRNG by repeatedly pull entropy from available entropy sources\n",
+				drng_type);
 		}
 
 		esdm_fill_seed_buffer(
@@ -459,7 +472,7 @@ static uint32_t esdm_drng_seed_es_nolock(struct esdm_drng *drng, bool init_ops,
 		if (init_ops)
 			esdm_init_ops(&collected_seedbuf);
 
-		/*
+	/*
 	 * Emergency reseeding: If we reached the min seed threshold now
 	 * multiple times but never reached fully seeded level and we collect
 	 * entropy, keep doing it until we reached fully seeded level for
@@ -500,12 +513,13 @@ static void esdm_drng_seed(struct esdm_drng *drng)
 
 static void esdm_drng_seed_work_one(struct esdm_drng *drng, uint32_t node)
 {
-	logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-	       "reseed triggered by system events for DRNG on node %d\n", node);
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_DRNG,
+		    "reseed triggered by system events for DRNG on node %d\n",
+		    node);
 	esdm_drng_seed(drng);
 	if (drng->fully_seeded) {
 		/* Prevent reseed storm */
-		drng->last_seeded += node * 60;
+		drng->last_seeded.tv_sec += node * 60;
 	}
 }
 
@@ -605,8 +619,8 @@ void esdm_drng_force_reseed(void)
 	    (atomic_read_u32(&esdm_drng_init.requests_since_fully_seeded) >
 	     ESDM_DRNG_RESEED_THRESH)) {
 		esdm_drng_init.force_reseed = esdm_drng_init.fully_seeded;
-		logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-		       "force reseed of initial DRNG\n");
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_DRNG,
+			    "force reseed of initial DRNG\n");
 		goto out;
 	}
 
@@ -617,8 +631,8 @@ void esdm_drng_force_reseed(void)
 			continue;
 
 		drng->force_reseed = drng->fully_seeded;
-		logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-		       "force reseed of DRNG on CPU %u\n", node);
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_DRNG,
+			    "force reseed of DRNG on CPU %u\n", node);
 	}
 
 	esdm_drng_atomic_force_reseed();
@@ -629,9 +643,11 @@ out:
 
 static bool esdm_drng_must_reseed(struct esdm_drng *drng)
 {
+	struct timespec check_time = drng->last_seeded;
+
+	check_time.tv_sec += esdm_drng_reseed_max_time;
 	return (atomic_dec_and_test(&drng->requests) || drng->force_reseed ||
-		esdm_time_after_now(drng->last_seeded +
-				    esdm_drng_reseed_max_time));
+		esdm_time_after_now(&check_time));
 }
 
 /**
@@ -742,9 +758,10 @@ static ssize_t esdm_drng_get(struct esdm_drng *drng, uint8_t *outbuf,
 						   outbuf + processed, todo);
 		mutex_w_unlock(&drng->lock);
 		if (ret <= 0) {
-			logger(LOGGER_WARN, LOGGER_C_DRNG,
-			       "getting random data from DRNG failed (%zd)\n",
-			       ret);
+			esdm_logger(
+				LOGGER_WARN, LOGGER_C_DRNG,
+				"getting random data from DRNG failed (%zd)\n",
+				ret);
 			return -EFAULT;
 		}
 		processed += ret;
@@ -770,18 +787,21 @@ static ssize_t esdm_drng_get_sleep(uint8_t *outbuf, size_t outbuflen, bool pr)
 	ssize_t ret;
 
 	if (pr) {
-		logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-		       "Using prediction resistance DRNG instance to service generate request\n");
+		esdm_logger(
+			LOGGER_DEBUG, LOGGER_C_DRNG,
+			"Using prediction resistance DRNG instance to service generate request\n");
 		drng = &esdm_drng_pr;
 	} else if (esdm_drng && esdm_drng[node] &&
 		   esdm_drng[node]->fully_seeded) {
-		logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-		       "Using DRNG instance on node %u to service generate request\n",
-		       node);
+		esdm_logger(
+			LOGGER_DEBUG, LOGGER_C_DRNG,
+			"Using DRNG instance on node %u to service generate request\n",
+			node);
 		drng = esdm_drng[node];
 	} else {
-		logger(LOGGER_DEBUG, LOGGER_C_DRNG,
-		       "Using DRNG instance on node 0 to service generate request\n");
+		esdm_logger(
+			LOGGER_DEBUG, LOGGER_C_DRNG,
+			"Using DRNG instance on node 0 to service generate request\n");
 	}
 
 	CKINT(esdm_drng_mgr_initialize());
@@ -881,6 +901,20 @@ static int esdm_drng_sleep_while_nonoperational(unsigned int nonblock)
 	return 0;
 }
 
+static int esdm_drng_sleep_while_nonoperational_timeout(struct timespec *ts)
+{
+	int ret = 0;
+
+	esdm_force_fully_seeded();
+	if (esdm_state_operational())
+		return 0;
+	thread_timedwait_event(&esdm_init_wait,
+			       esdm_state_operational() &&
+					!atomic_read(&esdm_drng_mgr_terminate),
+			       ts);
+	return ret;
+}
+
 static int esdm_drng_sleep_while_non_min_seeded(unsigned int nonblock)
 {
 	esdm_force_fully_seeded();
@@ -898,6 +932,7 @@ DSO_PUBLIC
 ssize_t esdm_get_seed(uint64_t *buf, size_t nbytes,
 		      enum esdm_get_seed_flags flags)
 {
+	static DEFINE_MUTEX_W_UNLOCKED(esdm_get_seed_lock);
 	struct entropy_buf *eb = (struct entropy_buf *)(buf + 2);
 	uint64_t buflen = sizeof(struct entropy_buf) + 2 * sizeof(uint64_t);
 	uint64_t collected_bits = 0;
@@ -913,10 +948,15 @@ ssize_t esdm_get_seed(uint64_t *buf, size_t nbytes,
 	if (nbytes < buflen)
 		return -EMSGSIZE;
 
-	ret = esdm_drng_sleep_while_not_all_nodes_seeded(
-		flags & ESDM_GET_SEED_NONBLOCK);
-	if (ret < 0)
-		return ret;
+	/*
+	 * We only allow ONE caller at any time to prevent a DoS on the RPC
+	 * interface considering this is a slow operation.
+	 */
+	if (!mutex_w_trylock(&esdm_get_seed_lock))
+		return -EAGAIN;
+
+	CKINT(esdm_drng_sleep_while_not_all_nodes_seeded(
+		flags & ESDM_GET_SEED_NONBLOCK));
 
 	/* Try to get the pool lock and sleep on it to get it. */
 	esdm_pool_lock();
@@ -924,7 +964,8 @@ ssize_t esdm_get_seed(uint64_t *buf, size_t nbytes,
 	/* If an ESDM DRNG becomes unseeded, give this DRNG precedence. */
 	if (!esdm_pool_all_nodes_seeded_get()) {
 		esdm_pool_unlock();
-		return 0;
+		buflen = 0;
+		goto out;
 	}
 
 	/*
@@ -958,14 +999,30 @@ ssize_t esdm_get_seed(uint64_t *buf, size_t nbytes,
 	/* Write collected entropy size into second word */
 	buf[1] = collected_bits;
 
-	return (ssize_t)buflen;
+out:
+	mutex_w_unlock(&esdm_get_seed_lock);
+	return ret ? ret : (ssize_t)buflen;
 }
 
 DSO_PUBLIC
 ssize_t esdm_get_random_bytes_pr(uint8_t *buf, size_t nbytes)
 {
+	static DEFINE_MUTEX_W_UNLOCKED(esdm_pr_lock);
+	ssize_t ret;
+
 	esdm_drng_sleep_while_nonoperational(0);
-	return esdm_drng_get_sleep(buf, (uint32_t)nbytes, true);
+
+	/*
+	 * We only allow one call in flight which is a precaution that
+	 * a caller cannot flood the RPC lines with requests to the slow
+	 * PR DRNG and cause a denial of service to the others.
+	 */
+	if (!mutex_w_trylock(&esdm_pr_lock))
+		return -EAGAIN;
+
+	ret = esdm_drng_get_sleep(buf, (uint32_t)nbytes, true);
+	mutex_w_unlock(&esdm_pr_lock);
+	return ret;
 }
 
 DSO_PUBLIC
@@ -995,6 +1052,18 @@ ssize_t esdm_get_random_bytes_full(uint8_t *buf, size_t nbytes)
 	return esdm_drng_get_sleep(buf, (uint32_t)nbytes, false);
 }
 
+DSO_PUBLIC
+ssize_t esdm_get_random_bytes_full_timeout(uint8_t *buf, size_t nbytes,
+					   struct timespec *ts)
+{
+	int ret = esdm_drng_sleep_while_nonoperational_timeout(ts);
+
+	if (ret)
+		return ret;
+
+	return esdm_drng_get_sleep(buf, (uint32_t)nbytes, false);
+}
+
 DSO_PUBLIC
 ssize_t esdm_get_random_bytes_min(uint8_t *buf, size_t nbytes)
 {
