@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2022 - 2023, Stephan Mueller <smueller@chronox.de>
+ * Copyright (C) 2022 - 2024, Stephan Mueller <smueller@chronox.de>
  *
  * License: see LICENSE file in root directory
  *
@@ -37,7 +37,7 @@
 #include "esdm_rpc_service.h"
 #include "helper.h"
 #include "linux_support.h"
-#include "logger.h"
+#include "esdm_logger.h"
 #include "math_helper.h"
 #include "memset_secure.h"
 #include "mutex.h"
@@ -59,7 +59,6 @@ static char *mount_dst = NULL;
  ******************************************************************************/
 
 static struct esdm_shm_status *esdm_cuse_shm_status = NULL;
-static int esdm_cuse_shmid = -1;
 static atomic_64_t next_fh = ATOMIC_64_INIT(1);
 
 static int esdm_cuse_shm_status_avail(void)
@@ -70,7 +69,7 @@ static int esdm_cuse_shm_status_avail(void)
 
 	if (ret && !initialized) {
 		initialized = 1;
-		logger_status(
+		esdm_logger_status(
 			LOGGER_C_CUSE,
 			"CUSE client started detected ESDM server with properties:\n%s\n",
 			esdm_cuse_shm_status->info);
@@ -89,6 +88,7 @@ static void esdm_cuse_shm_status_close_shm(void)
 
 static int esdm_cuse_shm_status_create_shm(void)
 {
+	int esdm_cuse_shmid = -1;
 	int errsv, create_shm;
 	void *tmp;
 
@@ -107,7 +107,6 @@ static int esdm_cuse_shm_status_create_shm(void)
 			esdm_cuse_shm_status_close_shm();
 			if (esdm_cuse_shmid >= 0) {
 				shmctl(esdm_cuse_shmid, IPC_RMID, NULL);
-				esdm_cuse_shmid = -1;
 			}
 			return -errsv;
 		}
@@ -131,27 +130,28 @@ static int esdm_cuse_shm_status_create_shm(void)
 					       S_IROTH);
 			if (esdm_cuse_shmid < 0) {
 				errsv = errno;
-				logger(LOGGER_ERR, LOGGER_C_ANY,
-				       "ESDM shared memory segment creation failed: %s\n",
-				       strerror(errsv));
+				esdm_logger(
+					LOGGER_ERR, LOGGER_C_ANY,
+					"ESDM shared memory segment creation failed: %s\n",
+					strerror(errsv));
 				return -errsv;
 			}
 		}
 	}
 
-	tmp = shmat(esdm_cuse_shmid, NULL, SHM_RDONLY);
+	tmp = shmat(esdm_cuse_shmid, NULL, 0);
 	if (tmp == (void *)-1) {
 		errsv = errno;
-		logger(LOGGER_ERR, LOGGER_C_CUSE,
-		       "Attaching to shared memory segment failed: %s\n",
-		       strerror(errsv));
+		esdm_logger(LOGGER_ERR, LOGGER_C_CUSE,
+			    "Attaching to shared memory segment failed: %s\n",
+			    strerror(errsv));
 		esdm_cuse_shm_status_close_shm();
 		return -errsv;
 	}
 	esdm_cuse_shm_status = tmp;
 
-	logger(LOGGER_DEBUG, LOGGER_C_CUSE,
-	       "ESDM shared memory segment successfully attached to\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_CUSE,
+		    "ESDM shared memory segment successfully attached to\n");
 
 	return 0;
 }
@@ -160,6 +160,7 @@ static int esdm_cuse_shm_status_create_shm(void)
  * Semaphore for shared memory segment
  ******************************************************************************/
 
+static atomic_bool_t esdm_cuse_poll_thread_shutdown = ATOMIC_BOOL_INIT(false);
 static sem_t *esdm_cuse_semid = SEM_FAILED;
 static const char *esdm_sem_name = NULL;
 
@@ -168,7 +169,8 @@ static void esdm_cuse_shm_status_down(void)
 	struct timespec ts = { .tv_sec = 1, .tv_nsec = 0 };
 
 	if (esdm_cuse_semid == SEM_FAILED) {
-		logger(LOGGER_ERR, LOGGER_C_CUSE, "Cannot use semaphore\n");
+		esdm_logger(LOGGER_ERR, LOGGER_C_CUSE,
+			    "Cannot use semaphore\n");
 		return;
 	}
 
@@ -176,8 +178,13 @@ static void esdm_cuse_shm_status_down(void)
 	while (!esdm_cuse_shm_status_avail())
 		nanosleep(&ts, NULL);
 
+	/* The server is terminating, do not block any more */
+	if (atomic_bool_read(&esdm_cuse_poll_thread_shutdown))
+		return;
+
 	if (sem_wait(esdm_cuse_semid))
-		logger(LOGGER_ERR, LOGGER_C_CUSE, "Cannot use semaphore\n");
+		esdm_logger(LOGGER_ERR, LOGGER_C_CUSE,
+			    "Cannot use semaphore\n");
 }
 
 static void esdm_cuse_shm_status_close_sem(void)
@@ -209,16 +216,16 @@ static int esdm_cuse_shm_status_create_sem(void)
 		}
 	}
 
-	logger(LOGGER_DEBUG, LOGGER_C_CUSE,
-	       "ESDM change indicator semaphore initialized\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_CUSE,
+		    "ESDM change indicator semaphore initialized\n");
 
 	return 0;
 
 err:
 	errsv = errno;
-	logger(LOGGER_ERR, LOGGER_C_ANY,
-	       "ESDM change indicator semaphore creation failed: %s\n",
-	       strerror(errsv));
+	esdm_logger(LOGGER_ERR, LOGGER_C_ANY,
+		    "ESDM change indicator semaphore creation failed: %s\n",
+		    strerror(errsv));
 	return -errsv;
 }
 
@@ -226,14 +233,14 @@ err:
  * Signal handler
  ******************************************************************************/
 
-static bool esdm_cuse_poll_thread_shutdown = false;
 static DECLARE_WAIT_QUEUE(esdm_cuse_poll_checker_wait);
 
 static void esdm_cuse_term(void)
 {
-	esdm_cuse_poll_thread_shutdown = true;
+	atomic_bool_set(&esdm_cuse_poll_thread_shutdown, true);
 	thread_wake_all(&esdm_cuse_poll_checker_wait);
-	sem_post(esdm_cuse_semid);
+	if (esdm_cuse_semid != SEM_FAILED)
+		sem_post(esdm_cuse_semid);
 
 	thread_stop_spawning();
 
@@ -256,7 +263,7 @@ static void esdm_cuse_term(void)
 /* terminate the daemon cleanly */
 static void esdm_cuse_sig_handler(int sig)
 {
-	logger(LOGGER_DEBUG, LOGGER_C_CUSE, "Received signal %d\n", sig);
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_CUSE, "Received signal %d\n", sig);
 	esdm_cuse_term();
 
 	signal(SIGABRT, SIG_DFL);
@@ -288,8 +295,8 @@ static void esdm_cuse_sig_handler(int sig)
 
 static int esdm_cuse_install_sig_handler(void)
 {
-	logger(LOGGER_DEBUG, LOGGER_C_CUSE,
-	       "Install termination signal handler\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_CUSE,
+		    "Install termination signal handler\n");
 
 	/* Catch all termination signals to ensure the bind mount is removed */
 	signal(SIGABRT, esdm_cuse_sig_handler);
@@ -346,7 +353,7 @@ static bool esdm_cuse_fips_enabled(void)
 					/* FIPS support not enabled in kernel */
 					return 0;
 				} else {
-					logger(LOGGER_ERR, LOGGER_C_CUSE,
+					esdm_logger(LOGGER_ERR, LOGGER_C_CUSE,
 						"FIPS: Cannot open fips_enabled file: %s\n",
 						strerror(errno));
 					return -EIO;
@@ -356,7 +363,7 @@ static bool esdm_cuse_fips_enabled(void)
 			n = fread((void *)fipsflag, 1, 1, fipsfile);
 			fclose(fipsfile);
 			if (n != 1) {
-				logger(LOGGER_ERR, LOGGER_C_CUSE,
+				esdm_logger(LOGGER_ERR, LOGGER_C_CUSE,
 				       "FIPS: Cannot read FIPS flag\n");
 				return false;
 			}
@@ -392,11 +399,12 @@ static bool esdm_cuse_client_privileged(fuse_req_t req)
 	 * MUST NOT run in a PID or user namespace.
 	 */
 	if (ctx->uid == 0) {
-		logger(LOGGER_DEBUG, LOGGER_C_CUSE, "CUSE caller privileged\n");
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_CUSE,
+			    "CUSE caller privileged\n");
 		return true;
 	}
 
-	logger(LOGGER_DEBUG, LOGGER_C_CUSE, "CUSE caller unprivileged\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_CUSE, "CUSE caller unprivileged\n");
 	return false;
 }
 
@@ -452,30 +460,38 @@ void esdm_cuse_read_internal(fuse_req_t req, size_t size, off_t off,
 			     struct fuse_file_info *fi, get_func_t get,
 			     int fallback_fd)
 {
-	uint8_t tmpbuf[ESDM_RPC_MAX_DATA];
-	size_t cleansize = min_size(sizeof(tmpbuf), size);
+	uint8_t tmpbuf_s[64], *tmpbuf_p = tmpbuf_s, *tmpbuf = NULL;
+	size_t read_bytes = 0;
 	ssize_t ret = 0;
 
 	(void)off;
 
+	/*
+	 * size is limited by fuse to its maximum request size, mostly
+	 * 131072 byte
+	 */
+	if (size > sizeof(tmpbuf_s)) {
+		tmpbuf = calloc(1, size);
+		CKNULL(tmpbuf, -ENOMEM);
+		tmpbuf_p = tmpbuf;
+	}
+
 	fallback_fd = esdm_test_fallback_fd(fallback_fd);
 
 	if (fi->flags & O_SYNC)
 		get = esdm_rpcc_get_random_bytes_pr_int;
 
-	if (size > sizeof(tmpbuf)) {
-		logger(LOGGER_ERR, LOGGER_C_CUSE,
-		       "Due to FUSE limitation, the maximum request size is %zu\n",
-		       sizeof(tmpbuf));
-		size = sizeof(tmpbuf);
-	}
-
-	//	while (size)
-	{
-		size_t todo = min_size(sizeof(tmpbuf), size);
+	/*
+	 * fuse automatically chunks requests, e.g. for a 1MB read
+	 * multiple <= 131072 byte reads are typically performed, try to fill
+	 * them up.
+	 */
+	while (read_bytes < size) {
+		size_t todo =
+			min_size(ESDM_RPC_MAX_MSG_SIZE, size - read_bytes);
 
 		esdm_cuse_unpriv_call_start();
-		esdm_invoke(get(tmpbuf, todo, req));
+		esdm_invoke(get(tmpbuf_p + read_bytes, todo, req));
 		esdm_cuse_unpriv_call_end();
 
 		/*
@@ -486,40 +502,27 @@ void esdm_cuse_read_internal(fuse_req_t req, size_t size, off_t off,
 		 * appropriately.
 		 */
 		if (ret < 0 && fallback_fd > -1) {
-			logger(LOGGER_VERBOSE, LOGGER_C_CUSE,
-			       "Use fallback to provide data due to RPC error code %zd\n",
-			       ret);
-			ret = read(fallback_fd, tmpbuf, todo);
+			esdm_logger(
+				LOGGER_VERBOSE, LOGGER_C_CUSE,
+				"Use fallback to provide data due to RPC error code %zd\n",
+				ret);
+			ret = read(fallback_fd, tmpbuf_p + read_bytes, todo);
 		}
 
 		if (ret < 0)
 			goto out;
-		todo = (size_t)ret;
-
-		/*
-		 * This call segfaults when called a 2nd time because req is
-		 * freed. Thus, the while loop is currently disabled.
-		 *
-		 * Thus, we will return a short read here that the caller must
-		 * consider. Returning a short read is permissible in VFS and
-		 * thus it is no error to apply a short read here.
-		 *
-		 * The caller may accommodate that with a while loop around
-		 * its read system call. Another example when using dd is
-		 * the following command that must be used:
-		 *
-		 * dd if=/dev/esdm of=out count=1 bs=65550 iflag=fullblock
-		 */
-		ret = fuse_reply_buf(req, (const char *)tmpbuf, todo);
-
-		// 		if (ret < 0)
-		// 			goto out;
-		//
-		// 		size -= todo;
+		read_bytes += (size_t)ret;
 	}
+	ret = fuse_reply_buf(req, (const char *)tmpbuf_p, size);
 
 out:
-	memset_secure(tmpbuf, 0, cleansize);
+	if (tmpbuf) {
+		memset_secure(tmpbuf, 0, size);
+		free(tmpbuf);
+	} else {
+		memset_secure(tmpbuf_s, 0, sizeof(tmpbuf_s));
+	}
+
 	if (ret < 0)
 		fuse_reply_err(req, (int)-ret);
 }
@@ -536,29 +539,40 @@ void esdm_cuse_write_internal(fuse_req_t req, const char *buf, size_t size,
 
 	fallback_fd = esdm_test_fallback_fd(fallback_fd);
 
-	esdm_cuse_unpriv_call_start();
-	esdm_invoke(esdm_rpcc_write_data_int((const uint8_t *)buf, size, req));
-	esdm_cuse_unpriv_call_end();
-	if (ret == 0)
-		written = size;
+	while (written < size) {
+		size_t todo = min_size(ESDM_RPC_MAX_MSG_SIZE, size - written);
 
+		esdm_cuse_unpriv_call_start();
+		esdm_invoke(esdm_rpcc_write_data_int(
+			(const uint8_t *)buf + written, todo, req));
+		esdm_cuse_unpriv_call_end();
+		if (ret == 0)
+			written += todo;
+		else
+			goto err;
+	}
+
+	goto out;
+
+err:
 	/*
 	 * If call to the ESDM server failed, let us fall back to the
-	 * fallback file descriptor. Yet, we do not cover for short
-	 * reads as this entire CUSE handling is prone to short reads
-	 * as outlined below. Thus, the caller needs to handle this
-	 * appropriately.
+	 * fallback file descriptor. It writes the entire buffer into the
+	 * fallback as we do not consider a mix-n-match of data written to
+	 * ESDM and data written to the fallback as appropriate.
 	 */
 	if (ret < 0 && fallback_fd > -1) {
-		logger(LOGGER_VERBOSE, LOGGER_C_CUSE,
-		       "Use fallback to provide data due to RPC error code %zd\n",
-		       ret);
+		esdm_logger(
+			LOGGER_VERBOSE, LOGGER_C_CUSE,
+			"Use fallback to provide data due to RPC error code %zd\n",
+			ret);
 		do {
 			ret = write(fallback_fd, buf, size);
 			written += (size_t)ret;
 		} while (ret > 0 && written < size);
 	}
 
+out:
 	if (ret < 0)
 		fuse_reply_err(req, (int)-ret);
 	else
@@ -800,50 +814,64 @@ struct esdm_cuse_poll {
 static struct esdm_cuse_poll esdm_cuse_polls[ESDM_CUSE_MAX_PH];
 static DEFINE_MUTEX_W_UNLOCKED(esdm_cuse_ph_lock);
 
-static void esdm_cuse_set_pollmask(unsigned int request_events,
-				   unsigned int *outmask)
+static void esdm_cuse_get_pollmask(unsigned int *outmask)
 {
+#define ESDM_POLL_READER (POLLIN | POLLRDNORM)
+#define ESDM_POLL_WRITER (POLLOUT | POLLWRNORM)
+
 	*outmask = 0;
 
 	if (atomic_bool_read(&esdm_cuse_shm_status->operational))
-		*outmask |= POLLIN | POLLRDNORM;
+		*outmask |= ESDM_POLL_READER;
 	if (atomic_bool_read(&esdm_cuse_shm_status->need_entropy))
-		*outmask |= POLLOUT | POLLWRNORM;
+		*outmask |= ESDM_POLL_WRITER;
 
+	/* Simply wake the poller no matter what it waits for. */
+	if (atomic_bool_read(&esdm_cuse_shm_status->suspend_trigger)) {
+		atomic_bool_set(&esdm_cuse_shm_status->suspend_trigger, false);
+		*outmask |= ESDM_POLL_READER | ESDM_POLL_WRITER;
+	}
+
+	/* Simply wake the poller no matter what it waits for. */
+	if (atomic_bool_read(&esdm_cuse_poll_thread_shutdown))
+		*outmask |= ESDM_POLL_READER | ESDM_POLL_WRITER;
+}
+
+/* *outmask is already filled with output of esdm_cuse_get_pollmask */
+static void esdm_cuse_set_pollmask(unsigned int request_events,
+				   unsigned int *outmask)
+{
 	*outmask &= request_events;
 }
 
+/* when extending this function, always check and return error
+ * codes first, as they may originate from an interrupted poll/select
+ * which causes many bogus events if answered with notify_poll or reply_poll
+ */
 void esdm_cuse_poll(fuse_req_t req, struct fuse_file_info *fi,
 		    struct fuse_pollhandle *ph)
 {
 	unsigned int i, mask;
+	int err_code;
 
 	if (!fi->poll_events) {
-		fuse_reply_err(req, EINVAL);
-		return;
+		err_code = EINVAL;
+		goto err;
 	}
 
 	/*
 	 * Check current status and return it if it complies with requested
 	 * status.
 	 */
+	esdm_cuse_get_pollmask(&mask);
 	esdm_cuse_set_pollmask(fi->poll_events, &mask);
 	fuse_reply_poll(req, mask);
 
-	if (!ph)
-		return;
-
-	if (mask) {
-		fuse_notify_poll(ph);
-		fuse_pollhandle_destroy(ph);
-		return;
-	}
-
+	/* cleanup first, as we may have an interrupted poll/select */
 	mutex_w_lock(&esdm_cuse_ph_lock);
 	for (i = 0; i < ESDM_CUSE_MAX_PH; i++) {
 		if (esdm_cuse_polls[i].fh == fi->fh) {
 			if (esdm_cuse_polls[i].ph) {
-				fuse_notify_poll(esdm_cuse_polls[i].ph);
 				fuse_pollhandle_destroy(esdm_cuse_polls[i].ph);
 			}
 			esdm_cuse_polls[i].fh = 0;
@@ -854,6 +882,12 @@ void esdm_cuse_poll(fuse_req_t req, struct fuse_file_info *fi,
 		if (esdm_cuse_polls[i].ph)
 			continue;
 
+		if (mask) {
+			fuse_notify_poll(ph);
+			fuse_pollhandle_destroy(ph);
+			break;
+		}
+
 		esdm_cuse_polls[i].fh = fi->fh;
 		esdm_cuse_polls[i].ph = ph;
 		esdm_cuse_polls[i].poll_events = fi->poll_events;
@@ -861,14 +895,21 @@ void esdm_cuse_poll(fuse_req_t req, struct fuse_file_info *fi,
 	}
 	mutex_w_unlock(&esdm_cuse_ph_lock);
 
-	if (i == ESDM_CUSE_MAX_PH)
-		fuse_reply_err(req, EBUSY);
+	if (i == ESDM_CUSE_MAX_PH) {
+		err_code = EBUSY;
+		goto err;
+	}
+
+	return;
+
+err:
+	fuse_reply_err(req, err_code);
 }
 
 /* Poll checker handler executed in separate thread */
 static int esdm_cuse_poll_checker(void __unused *unused)
 {
-	unsigned int i, mask;
+	unsigned int i;
 
 	thread_set_name(cuse_poll, 0);
 
@@ -880,12 +921,22 @@ static int esdm_cuse_poll_checker(void __unused *unused)
 	}
 	thread_wake_all(&esdm_cuse_poll_checker_wait);
 
-	while (!esdm_cuse_poll_thread_shutdown) {
+	while (!atomic_bool_read(&esdm_cuse_poll_thread_shutdown)) {
+		unsigned int sysmask, mask;
+		bool sysmask_set = false;
+
 		mutex_w_lock(&esdm_cuse_ph_lock);
 		for (i = 0; i < ESDM_CUSE_MAX_PH; i++) {
 			if (!esdm_cuse_polls[i].ph)
 				continue;
 
+			/* Get the mask once for this loop */
+			if (!sysmask_set) {
+				esdm_cuse_get_pollmask(&sysmask);
+				sysmask_set = true;
+			}
+
+			mask = sysmask;
 			esdm_cuse_set_pollmask(esdm_cuse_polls[i].poll_events,
 					       &mask);
 
@@ -939,9 +990,10 @@ void esdm_cuse_init_done(void *userdata)
 	if (mount_src) {
 		if (chmod(mount_src, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP |
 					     S_IROTH | S_IWOTH) < 0) {
-			logger(LOGGER_ERR, LOGGER_C_CUSE,
-			       "Changing permissions to world-writeable failed: %s",
-			       strerror(errno));
+			esdm_logger(
+				LOGGER_ERR, LOGGER_C_CUSE,
+				"Changing permissions to world-writeable failed: %s",
+				strerror(errno));
 		}
 	}
 
@@ -1052,12 +1104,13 @@ int main_common(const char *_devname, const char *target, const char *semname,
 
 	if (fuse_opt_parse(&args, &param, esdm_cuse_opts,
 			   esdm_cuse_process_arg)) {
-		logger(LOGGER_ERR, LOGGER_C_CUSE, "failed to parse option\n");
+		esdm_logger(LOGGER_ERR, LOGGER_C_CUSE,
+			    "failed to parse option\n");
 		free(param.dev_name);
 		goto out;
 	}
 
-	logger_set_verbosity(param.verbosity);
+	esdm_logger_set_verbosity(param.verbosity);
 
 	esdm_test_disable_fallback(param.disable_fallback);
 
