@@ -1,6 +1,6 @@
 /* RPC Client: Connection handler to server
  *
- * Copyright (C) 2022 - 2023, Stephan Mueller <smueller@chronox.de>
+ * Copyright (C) 2022 - 2024, Stephan Mueller <smueller@chronox.de>
  *
  * License: see LICENSE file in root directory
  *
@@ -33,9 +33,10 @@
 #include "conv_be_le.h"
 #include "esdm_rpc_client_internal.h"
 #include "esdm_rpc_protocol.h"
+#include "esdm_rpc_protocol_helper.h"
 #include "esdm_rpc_service.h"
 #include "helper.h"
-#include "logger.h"
+#include "esdm_logger.h"
 #include "math_helper.h"
 #include "memset_secure.h"
 #include "ptr_err.h"
@@ -65,6 +66,9 @@ static void esdm_fini_proto_service(esdm_rpc_client_connection_t *rpc_conn)
 		protobuf_c_service_destroy(service);
 		service->descriptor = NULL;
 	}
+
+	mutex_w_destroy(&rpc_conn->lock);
+	mutex_w_destroy(&rpc_conn->ref_cnt);
 }
 
 static int esdm_connect_proto_service(esdm_rpc_client_connection_t *rpc_conn)
@@ -93,16 +97,17 @@ static int esdm_connect_proto_service(esdm_rpc_client_connection_t *rpc_conn)
 		errsv = errno;
 
 		if (errsv == ENOENT) {
-			logger(LOGGER_DEBUG, LOGGER_C_RPC,
-			       "ESDM server interface %s not available\n",
-			       socketname);
+			esdm_logger(LOGGER_DEBUG, LOGGER_C_RPC,
+				    "ESDM server interface %s not available\n",
+				    socketname);
 		}
 
 		return -errsv;
 	}
 
-	logger(LOGGER_DEBUG, LOGGER_C_RPC,
-	       "Attempting to access ESDM server interface %s\n", socketname);
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_RPC,
+		    "Attempting to access ESDM server interface %s\n",
+		    socketname);
 
 	/* Connect to the Unix domain socket */
 	addr.sun_family = AF_UNIX;
@@ -111,8 +116,8 @@ static int esdm_connect_proto_service(esdm_rpc_client_connection_t *rpc_conn)
 	if (rpc_conn->fd < 0) {
 		errsv = errno;
 
-		logger(LOGGER_ERR, LOGGER_C_RPC, "Error creating socket: %s\n",
-		       strerror(errsv));
+		esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+			    "Error creating socket: %s\n", strerror(errsv));
 		return -errsv;
 	}
 
@@ -123,9 +128,9 @@ static int esdm_connect_proto_service(esdm_rpc_client_connection_t *rpc_conn)
 		       sizeof(tv)) < 0) {
 		errsv = errno;
 
-		logger(LOGGER_ERR, LOGGER_C_RPC,
-		       "Error setting timeout on socket: %s\n",
-		       strerror(errsv));
+		esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+			    "Error setting timeout on socket: %s\n",
+			    strerror(errsv));
 		return -errsv;
 	}
 
@@ -138,9 +143,9 @@ static int esdm_connect_proto_service(esdm_rpc_client_connection_t *rpc_conn)
 			    sizeof(addr)) < 0) {
 			errsv = errno;
 
-			logger(LOGGER_ERR, LOGGER_C_RPC,
-			       "Error connecting socket: %s\n",
-			       strerror(errsv));
+			esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+				    "Error connecting socket: %s\n",
+				    strerror(errsv));
 			attempts++;
 		} else {
 			errsv = 0;
@@ -149,16 +154,16 @@ static int esdm_connect_proto_service(esdm_rpc_client_connection_t *rpc_conn)
 		 (errsv == EAGAIN || errsv == ECONNREFUSED || errsv == EINTR));
 
 	if (errsv) {
-		logger(LOGGER_ERR, LOGGER_C_RPC,
-		       "Connection attempt using socket %s failed\n",
-		       socketname);
+		esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+			    "Connection attempt using socket %s failed\n",
+			    socketname);
 	}
 
 	return -errsv;
 }
 
-static int esdm_rpc_client_write_data(esdm_rpc_client_connection_t *rpc_conn,
-				      const uint8_t *data, size_t len)
+static int esdm_rpc_client_write_data_fd(esdm_rpc_client_connection_t *rpc_conn,
+					 const uint8_t *data, size_t len)
 {
 	size_t written = 0;
 	ssize_t ret;
@@ -188,8 +193,9 @@ static int esdm_rpc_client_write_data(esdm_rpc_client_connection_t *rpc_conn,
 			}
 
 			if (errsv == EPIPE) {
-				logger(LOGGER_DEBUG, LOGGER_C_RPC,
-				       "Connection to server needs to be re-established\n");
+				esdm_logger(
+					LOGGER_DEBUG, LOGGER_C_RPC,
+					"Connection to server needs to be re-established\n");
 
 				int rc = esdm_connect_proto_service(rpc_conn);
 				if (rc)
@@ -197,9 +203,10 @@ static int esdm_rpc_client_write_data(esdm_rpc_client_connection_t *rpc_conn,
 				continue;
 			}
 
-			logger(LOGGER_ERR, LOGGER_C_RPC,
-			       "Writting of data to file descriptor %d failed: %s\n",
-			       rpc_conn->fd, strerror(errsv));
+			esdm_logger(
+				LOGGER_ERR, LOGGER_C_RPC,
+				"Writting of data to file descriptor %d failed: %s\n",
+				rpc_conn->fd, strerror(errsv));
 
 			return -errsv;
 		}
@@ -211,23 +218,107 @@ static int esdm_rpc_client_write_data(esdm_rpc_client_connection_t *rpc_conn,
 		len -= (size_t)ret;
 	} while (written < len);
 
-	logger(LOGGER_DEBUG2, LOGGER_C_ANY, "%zu bytes written\n", len);
+	esdm_logger(LOGGER_DEBUG2, LOGGER_C_ANY, "%zu bytes written\n", len);
 
 	return 0;
 }
 
+#ifdef ESDM_RPCC_BUF_WRITE
+
+/*
+ * Implementation of packing data and sending it out. Properties:
+ *
+ * - one call to write data out to the file descriptor
+ *
+ * - one more copy of entire data required to linearize all data
+ */
+static int esdm_rpc_client_pack(const ProtobufCMessage *message,
+				unsigned int method_index,
+				esdm_rpc_client_connection_t *rpc_conn)
+{
+#define ESDM_RPCC_BUF_WRITE_HEADER_SZ (sizeof(struct esdm_rpc_proto_cs_header))
+
+	size_t message_length;
+	int ret;
+	uint8_t *data_buf_alloc = NULL;
+	uint8_t *data_buf;
+	struct esdm_rpc_proto_cs_header *cs_header;
+	struct esdm_rpc_write_data_buf tmp = {
+		.dst_written = 0,
+	};
+
+	tmp.base.append = esdm_rpc_append_data;
+
+	message_length = protobuf_c_message_get_packed_size(message);
+	if (message_length > ESDM_RPC_MAX_MSG_SIZE) {
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_ANY,
+			    "Unexpected message length: %zu\n", message_length);
+		return -EFAULT;
+	}
+
+	data_buf_alloc = malloc(ESDM_RPCC_BUF_WRITE_HEADER_SZ + message_length +
+				sizeof(uint64_t) - 1);
+	CKNULL(data_buf_alloc, -ENOMEM);
+
+	data_buf = ALIGN_PTR_8(data_buf_alloc, sizeof(uint64_t));
+	tmp.dst_buf = (data_buf + ESDM_RPCC_BUF_WRITE_HEADER_SZ);
+
+	cs_header = (struct esdm_rpc_proto_cs_header *)data_buf;
+	cs_header->method_index = le_bswap32(method_index);
+	cs_header->message_length = le_bswap32(message_length);
+	cs_header->request_id = le_bswap32(0);
+
+	esdm_logger(
+		LOGGER_DEBUG, LOGGER_C_RPC,
+		"Client sending: message length %u, message index %u, request ID %u\n",
+		cs_header->message_length, cs_header->method_index,
+		cs_header->request_id);
+
+	if (protobuf_c_message_pack_to_buffer(message, &tmp.base) !=
+	    message_length) {
+		esdm_logger(LOGGER_VERBOSE, LOGGER_C_RPC,
+			    "Short write of data to file descriptor \n");
+		ret = -EFAULT;
+		goto out;
+	}
+
+	CKINT_LOG(esdm_rpc_client_write_data_fd(rpc_conn, data_buf,
+						ESDM_RPCC_BUF_WRITE_HEADER_SZ +
+							message_length),
+		  "Submission of message data failed with error %d\n", ret);
+
+out:
+	/*
+	 * Zeroization not needed, as data will go out through unprotected
+	 * channel anyway. If the data is not already protected, there is a
+	 * bigger problem.
+	 */
+	if (data_buf_alloc)
+		free(data_buf_alloc);
+	return ret;
+}
+
+#else /* ESDM_RPCC_BUF_WRITE */
+
 static void esdm_rpc_client_append_data(ProtobufCBuffer *buffer, size_t len,
 					const uint8_t *data)
 {
 	struct esdm_rpcc_write_buf *buf = (struct esdm_rpcc_write_buf *)buffer;
-	int ret = esdm_rpc_client_write_data(buf->rpc_conn, data, len);
+	int ret = esdm_rpc_client_write_data_fd(buf->rpc_conn, data, len);
 
 	if (ret < 0)
-		logger(LOGGER_ERR, LOGGER_C_RPC,
-		       "Submission of payload data failed with error %d\n",
-		       ret);
+		esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+			    "Submission of payload data failed with error %d\n",
+			    ret);
 }
 
+/*
+ * Implementation of packing data and sending it out. Properties:
+ *
+ * - multiple calls to write data out to the file descriptor
+ *
+ * - no additional memory required
+ */
 static int esdm_rpc_client_pack(const ProtobufCMessage *message,
 				unsigned int method_index,
 				esdm_rpc_client_connection_t *rpc_conn)
@@ -245,19 +336,20 @@ static int esdm_rpc_client_pack(const ProtobufCMessage *message,
 	cs_header.message_length = le_bswap32(message_length);
 	cs_header.request_id = le_bswap32(0);
 
-	logger(LOGGER_DEBUG, LOGGER_C_RPC,
-	       "Client sending: message length %u, message index %u, request ID %u\n",
-	       cs_header.message_length, cs_header.method_index,
-	       cs_header.request_id);
+	esdm_logger(
+		LOGGER_DEBUG, LOGGER_C_RPC,
+		"Client sending: message length %u, message index %u, request ID %u\n",
+		cs_header.message_length, cs_header.method_index,
+		cs_header.request_id);
 
-	CKINT_LOG(esdm_rpc_client_write_data(rpc_conn, (uint8_t *)&cs_header,
-					     sizeof(cs_header)),
+	CKINT_LOG(esdm_rpc_client_write_data_fd(rpc_conn, (uint8_t *)&cs_header,
+						sizeof(cs_header)),
 		  "Submission of header data failed with error %d\n", ret);
 
 	if (protobuf_c_message_pack_to_buffer(message, &tmp.base) !=
 	    message_length) {
-		logger(LOGGER_VERBOSE, LOGGER_C_RPC,
-		       "Short write of data to file descriptor \n");
+		esdm_logger(LOGGER_VERBOSE, LOGGER_C_RPC,
+			    "Short write of data to file descriptor \n");
 		ret = -EFAULT;
 	}
 
@@ -265,6 +357,8 @@ out:
 	return ret;
 }
 
+#endif /* ESDM_RPCC_BUF_WRITE */
+
 static int
 esdm_rpc_client_read_handler(esdm_rpc_client_connection_t *rpc_conn,
 			     const ProtobufCMessageDescriptor *message_desc,
@@ -320,8 +414,8 @@ esdm_rpc_client_read_handler(esdm_rpc_client_connection_t *rpc_conn,
 			}
 
 			ret = -errno;
-			logger(LOGGER_DEBUG, LOGGER_C_RPC, "Read failed: %s\n",
-			       strerror(errno));
+			esdm_logger(LOGGER_DEBUG, LOGGER_C_RPC,
+				    "Read failed: %s\n", strerror(errno));
 			break;
 		}
 
@@ -347,10 +441,11 @@ esdm_rpc_client_read_handler(esdm_rpc_client_connection_t *rpc_conn,
 			header->method_index = le_bswap32(header->method_index);
 			header->request_id = le_bswap32(header->request_id);
 
-			logger(LOGGER_DEBUG, LOGGER_C_RPC,
-			       "Client received: server status %u, message length %u, message index %u, request ID %u\n",
-			       header->status_code, header->message_length,
-			       header->method_index, header->request_id);
+			esdm_logger(
+				LOGGER_DEBUG, LOGGER_C_RPC,
+				"Client received: server status %u, message length %u, message index %u, request ID %u\n",
+				header->status_code, header->message_length,
+				header->method_index, header->request_id);
 
 			/*
 			 * Truncate the buffer length if client specified
@@ -393,26 +488,28 @@ esdm_rpc_client_read_handler(esdm_rpc_client_connection_t *rpc_conn,
 			closure(msg, closure_data);
 			protobuf_c_message_free_unpacked(
 				msg, &esdm_rpc_client_allocator);
-			logger(LOGGER_DEBUG, LOGGER_C_RPC,
-			       "Data with length %u send to client closure handler\n",
-			       header->message_length);
+			esdm_logger(
+				LOGGER_DEBUG, LOGGER_C_RPC,
+				"Data with length %u send to client closure handler\n",
+				header->message_length);
 		} else {
-			logger(LOGGER_ERR, LOGGER_C_RPC,
-			       "Response message not found\n");
+			esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+				    "Response message not found\n");
 			msg = ERR_PTR(-EFAULT);
 			closure(msg, closure_data);
 		}
 	} else if (interrupted) {
 		ProtobufCMessage *msg;
 
-		logger(LOGGER_VERBOSE, LOGGER_C_RPC, "Request interrupted\n");
+		esdm_logger(LOGGER_VERBOSE, LOGGER_C_RPC,
+			    "Request interrupted\n");
 		msg = ERR_PTR(-EINTR);
 		closure(msg, closure_data);
 	} else {
 		ProtobufCMessage *msg;
 
-		logger(LOGGER_VERBOSE, LOGGER_C_RPC,
-		       "Server returned with an error\n");
+		esdm_logger(LOGGER_VERBOSE, LOGGER_C_RPC,
+			    "Server returned with an error\n");
 		msg = ERR_PTR(-EINTR);
 		closure(msg, closure_data);
 	}
@@ -437,7 +534,12 @@ static void esdm_client_invoke(ProtobufCService *service,
 	mutex_w_lock(&rpc_conn->lock);
 
 	do {
-		CKINT(esdm_connect_proto_service(rpc_conn));
+		/*
+		 * Connect to the server if we do not have a connection,
+		 * otherwise reuse the session.
+		 */
+		if (rpc_conn->fd == -1)
+			CKINT(esdm_connect_proto_service(rpc_conn));
 
 		/* Pack the protobuf-c data and send it over the wire */
 		CKINT_LOG(esdm_rpc_client_pack(input, method_index, rpc_conn),
@@ -516,16 +618,23 @@ static uint32_t esdm_rpcc_curr_node(void)
 	return (esdm_curr_node() % esdm_rpcc_max_nodes);
 }
 
-static void esdm_rpcc_fini_service(esdm_rpc_client_connection_t **rpc_conn)
+static void esdm_rpcc_fini_service(esdm_rpc_client_connection_t **rpc_conn,
+				   uint32_t *num)
 {
-	const struct timespec abstime = { .tv_sec = 1, .tv_nsec = 0 };
-	esdm_rpc_client_connection_t *rpc_conn_array = *rpc_conn;
-	esdm_rpc_client_connection_t *rpc_conn_p = rpc_conn_array;
-	uint32_t i, num_conn = esdm_rpcc_get_online_nodes();
+	struct timespec abstime;
+	esdm_rpc_client_connection_t *rpc_conn_array;
+	esdm_rpc_client_connection_t *rpc_conn_p;
+	uint32_t i, num_conn = *num;
+	int lock_res;
 
+	/* Atomic exchange */
+	*num = 0;
+	rpc_conn_array = __sync_lock_test_and_set(rpc_conn, NULL);
 	if (!rpc_conn_array)
 		return;
 
+	rpc_conn_p = rpc_conn_array;
+
 	esdm_test_shm_status_fini();
 
 	/* Tell everybody that the connection is about to terminate */
@@ -544,7 +653,10 @@ static void esdm_rpcc_fini_service(esdm_rpc_client_connection_t **rpc_conn)
 		 * we want to avoid a deadlock as the lock will not be released
 		 * by a killed thread.
 		 */
-		if (mutex_w_timedlock(&rpc_conn_p->ref_cnt, &abstime))
+		clock_gettime(CLOCK_MONOTONIC, &abstime);
+		abstime.tv_sec += 1;
+		lock_res = mutex_w_timedlock(&rpc_conn_p->ref_cnt, &abstime);
+		if (lock_res == 0 || lock_res == ETIMEDOUT)
 			mutex_w_unlock(&rpc_conn_p->ref_cnt);
 
 		/* Terminate the handle */
@@ -552,18 +664,61 @@ static void esdm_rpcc_fini_service(esdm_rpc_client_connection_t **rpc_conn)
 	}
 
 	free(rpc_conn_array);
-	*rpc_conn = NULL;
 }
 
 static int esdm_rpcc_init_service(const ProtobufCServiceDescriptor *descriptor,
 				  const char *socketname,
 				  esdm_rpcc_interrupt_func_t interrupt_func,
-				  esdm_rpc_client_connection_t **rpc_conn)
+				  esdm_rpc_client_connection_t **rpc_conn,
+				  uint32_t *num_conn)
 {
-	esdm_rpc_client_connection_t *tmp, *tmp_p;
+	esdm_rpc_client_connection_t *tmp = *rpc_conn, *tmp_p;
 	uint32_t i = 0, nodes = esdm_rpcc_get_online_nodes();
 	int ret = 0;
 
+	/*
+	 * It is a legitimate scenario that this function is called twice for
+	 * one connection as follows: if the libesdm_getrandom is preloaded, the
+	 * connection is already allocated. Now, the caller also wants to
+	 * establish a new connection, it cannot assume that libesdm_getrandom
+	 * is preloaded. Thus, it will unconditionally call the init function
+	 * as well. Thus catch this issue here and avoid double allocation.
+	 * Note, the esdm_rpcc_fini_service will ensure that there is also no
+	 * double free.
+	 */
+	if (tmp) {
+		/*
+		 * If the existing nodes are already sufficient, do not allocate
+		 * more.
+		 */
+		if (*num_conn >= nodes)
+			return 0;
+
+		/*
+		 * The caller wants more ESDM connections now - release all
+		 * connections to allocate them anew.
+		 *
+		 * It is inefficient to release the connections, free the memory
+		 * and allocate it anew. Yet, it is the hope that this code
+		 * is never called by assuming the user avoids such situations.
+		 *
+		 * But it is conceivable to have such situations:
+		 * libesdm_getrandom allocates memory for one connection. If
+		 * this library is preloaded, and the application itself wants
+		 * to allow a larger set of ESDM connections, we end up in this
+		 * code path.
+		 *
+		 * NOTE: This is only permissible when assuming that the
+		 * init_service call is done at the beginning of an application,
+		 * i.e. when there is no transaction running.
+		 */
+		for (i = 0, tmp_p = tmp; i < *num_conn; i++, tmp_p++)
+			esdm_fini_proto_service(tmp_p);
+		if (tmp)
+			free(tmp);
+		*num_conn = 0;
+	}
+
 	tmp = calloc(nodes, sizeof(*tmp));
 	CKNULL(tmp, -ENOMEM);
 
@@ -572,14 +727,19 @@ static int esdm_rpcc_init_service(const ProtobufCServiceDescriptor *descriptor,
 					      interrupt_func, tmp_p));
 	}
 
-	*rpc_conn = tmp;
-
-	logger(LOGGER_DEBUG, LOGGER_C_ANY,
-	       "Service supporting %u parallel requests for socket %s enabled\n",
-	       nodes, socketname);
-
 	CKINT(esdm_test_shm_status_init());
 
+	if (__sync_val_compare_and_swap(rpc_conn, NULL, tmp) != NULL) {
+		ret = -EAGAIN;
+		goto out;
+	}
+	*num_conn = nodes;
+
+	esdm_logger(
+		LOGGER_DEBUG, LOGGER_C_ANY,
+		"Service supporting %u parallel requests for socket %s enabled\n",
+		nodes, socketname);
+
 out:
 	if (ret) {
 		uint32_t j;
@@ -594,16 +754,19 @@ out:
 }
 
 static int esdm_rpcc_get_service(esdm_rpc_client_connection_t *rpc_conn_array,
+				 uint32_t num_conn,
 				 esdm_rpc_client_connection_t **ret_rpc_conn,
 				 void *int_data)
 {
 	esdm_rpc_client_connection_t *rpc_conn_p;
+	/* Protection against client programming errors */
+	uint32_t node = min_uint32(esdm_rpcc_curr_node(), num_conn);
 	int ret = 0;
 
 	CKNULL(rpc_conn_array, -EFAULT);
 	CKNULL(ret_rpc_conn, -EFAULT);
 
-	rpc_conn_p = rpc_conn_array + esdm_rpcc_curr_node();
+	rpc_conn_p = rpc_conn_array + node;
 
 	/*
 	 * Wait until the previous call completed - each connection handle is
@@ -640,12 +803,14 @@ static void esdm_rpcc_put_service(esdm_rpc_client_connection_t *rpc_conn)
  * Unprivileged connection
  ******************************************************************************/
 static esdm_rpc_client_connection_t *unpriv_rpc_conn = NULL;
+static uint32_t unpriv_rpc_conn_num = 0;
 
 DSO_PUBLIC
 int esdm_rpcc_get_unpriv_service(esdm_rpc_client_connection_t **rpc_conn,
 				 void *int_data)
 {
-	return esdm_rpcc_get_service(unpriv_rpc_conn, rpc_conn, int_data);
+	return esdm_rpcc_get_service(unpriv_rpc_conn, unpriv_rpc_conn_num,
+				     rpc_conn, int_data);
 }
 
 DSO_PUBLIC
@@ -659,25 +824,27 @@ int esdm_rpcc_init_unpriv_service(esdm_rpcc_interrupt_func_t interrupt_func)
 {
 	return esdm_rpcc_init_service(&unpriv_access__descriptor,
 				      ESDM_RPC_UNPRIV_SOCKET, interrupt_func,
-				      &unpriv_rpc_conn);
+				      &unpriv_rpc_conn, &unpriv_rpc_conn_num);
 }
 
 DSO_PUBLIC
 void esdm_rpcc_fini_unpriv_service(void)
 {
-	esdm_rpcc_fini_service(&unpriv_rpc_conn);
+	esdm_rpcc_fini_service(&unpriv_rpc_conn, &unpriv_rpc_conn_num);
 }
 
 /******************************************************************************
  * Privileged connection
  ******************************************************************************/
 static esdm_rpc_client_connection_t *priv_rpc_conn = NULL;
+static uint32_t priv_rpc_conn_num = 0;
 
 DSO_PUBLIC
 int esdm_rpcc_get_priv_service(esdm_rpc_client_connection_t **rpc_conn,
 			       void *int_data)
 {
-	return esdm_rpcc_get_service(priv_rpc_conn, rpc_conn, int_data);
+	return esdm_rpcc_get_service(priv_rpc_conn, priv_rpc_conn_num, rpc_conn,
+				     int_data);
 }
 
 DSO_PUBLIC
@@ -691,11 +858,11 @@ int esdm_rpcc_init_priv_service(esdm_rpcc_interrupt_func_t interrupt_func)
 {
 	return esdm_rpcc_init_service(&priv_access__descriptor,
 				      ESDM_RPC_PRIV_SOCKET, interrupt_func,
-				      &priv_rpc_conn);
+				      &priv_rpc_conn, &priv_rpc_conn_num);
 }
 
 DSO_PUBLIC
 void esdm_rpcc_fini_priv_service(void)
 {
-	esdm_rpcc_fini_service(&priv_rpc_conn);
+	esdm_rpcc_fini_service(&priv_rpc_conn, &priv_rpc_conn_num);
 }
