@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2022 - 2023, Stephan Mueller <smueller@chronox.de>
+ * Copyright (C) 2022 - 2024, Stephan Mueller <smueller@chronox.de>
  *
  * License: see LICENSE file in root directory
  *
@@ -28,6 +28,7 @@
 #include <stdio.h>
 #include <string.h>
 #include <sys/shm.h>
+#include <sys/select.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/un.h>
@@ -41,12 +42,13 @@
 #include "esdm.h"
 #include "esdm_config.h"
 #include "esdm_rpc_protocol.h"
+#include "esdm_rpc_protocol_helper.h"
 #include "esdm_rpc_server.h"
 #include "esdm_rpc_server_linux.h"
 #include "esdm_rpc_service.h"
 #include "helper.h"
 #include "linux_support.h"
-#include "logger.h"
+#include "esdm_logger.h"
 #include "memset_secure.h"
 #include "privileges.h"
 #include "ret_checkers.h"
@@ -131,20 +133,104 @@ static int esdm_rpcs_write_data(struct esdm_rpcs_connection *rpc_conn,
 		if (ret < 0) {
 			int errsv = errno;
 
-			logger(LOGGER_VERBOSE, LOGGER_C_RPC,
-			       "Writting of data to file descriptor %d failed: %s\n",
-			       rpc_conn->child_fd, strerror(errsv));
+			esdm_logger(
+				LOGGER_VERBOSE, LOGGER_C_RPC,
+				"Writting of data to file descriptor %d failed: %s\n",
+				rpc_conn->child_fd, strerror(errsv));
+
+			if (errsv == EPIPE) {
+				close(rpc_conn->child_fd);
+				rpc_conn->child_fd = -1;
+			}
+
 			return -errsv;
 		}
 
 		written += (size_t)ret;
 	} while (written < len);
 
-	logger(LOGGER_DEBUG2, LOGGER_C_ANY, "%zu bytes written\n", len);
+	esdm_logger(LOGGER_DEBUG2, LOGGER_C_ANY, "%zu bytes written\n", len);
 
 	return 0;
 }
 
+#ifdef ESDM_RPCS_BUF_WRITE
+
+/*
+ * Implementation of packing data and sending it out. Properties:
+ *
+ * - one call to write data out to the file descriptor
+ *
+ * - one more copy of entire data required to linearize all data
+ */
+static int esdm_rpcs_pack_internal(const ProtobufCMessage *message,
+				   struct esdm_rpcs_connection *rpc_conn)
+{
+#define ESDM_RPCS_BUF_WRITE_HEADER_SZ (sizeof(struct esdm_rpc_proto_sc_header))
+
+	size_t message_length;
+	int ret;
+	uint8_t *data_buf_alloc = NULL;
+	uint8_t *data_buf;
+	struct esdm_rpc_proto_sc_header *sc_header;
+	struct esdm_rpc_write_data_buf tmp = {
+		.dst_written = 0,
+	};
+
+	tmp.base.append = esdm_rpc_append_data;
+
+	message_length = protobuf_c_message_get_packed_size(message);
+	if (message_length > ESDM_RPC_MAX_MSG_SIZE) {
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_ANY,
+			    "Unexpected message length: %zu\n", message_length);
+		return -EFAULT;
+	}
+
+	data_buf_alloc = malloc(ESDM_RPCS_BUF_WRITE_HEADER_SZ + message_length +
+				sizeof(uint64_t) - 1);
+	CKNULL(data_buf_alloc, -ENOMEM);
+
+	data_buf = ALIGN_PTR_8(data_buf_alloc, sizeof(uint64_t));
+	tmp.dst_buf = (data_buf + ESDM_RPCS_BUF_WRITE_HEADER_SZ);
+
+	sc_header = (struct esdm_rpc_proto_sc_header *)data_buf;
+	sc_header->status_code = le_bswap32(PROTOBUF_C_RPC_STATUS_CODE_SUCCESS);
+	sc_header->method_index = le_bswap32(rpc_conn->method_index);
+	sc_header->message_length = le_bswap32(message_length);
+	sc_header->request_id = le_bswap32(rpc_conn->request_id);
+
+	esdm_logger(
+		LOGGER_DEBUG, LOGGER_C_RPC,
+		"Server sending: server status %u, message length %u, message index %u, request ID %u\n",
+		sc_header->status_code, sc_header->message_length,
+		sc_header->method_index, sc_header->request_id);
+
+	if (protobuf_c_message_pack_to_buffer(message, &tmp.base) !=
+	    message_length) {
+		esdm_logger(LOGGER_VERBOSE, LOGGER_C_RPC,
+			    "Short write of data to file descriptor\n");
+		ret = -EFAULT;
+		goto out;
+	}
+
+	CKINT_LOG(esdm_rpcs_write_data(rpc_conn, data_buf,
+				       ESDM_RPCS_BUF_WRITE_HEADER_SZ +
+					       message_length),
+		  "Submission of message data failed with error %d\n", ret);
+
+out:
+	/*
+	 * Zeroization not needed, as data will go out through unprotected
+	 * channel anyway. If the data is not already protected, there is a
+	 * bigger problem.
+	 */
+	if (data_buf_alloc)
+		free(data_buf_alloc);
+	return ret;
+}
+
+#else /* ESDM_RPCS_BUF_WRITE */
+
 /* Write out data from a ProtobufC buffer. */
 static void esdm_rpcs_append_data(ProtobufCBuffer *buffer, size_t len,
 				  const uint8_t *data)
@@ -153,19 +239,63 @@ static void esdm_rpcs_append_data(ProtobufCBuffer *buffer, size_t len,
 	int ret = esdm_rpcs_write_data(buf->rpc_conn, data, len);
 
 	if (ret < 0)
-		logger(LOGGER_ERR, LOGGER_C_RPC,
-		       "Submission of payload data failed with error %d\n",
-		       ret);
+		esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+			    "Submission of payload data failed with error %d\n",
+			    ret);
 }
 
+/*
+ * Implementation of packing data and sending it out. Properties:
+ *
+ * - multiple calls to write data out to the file descriptor
+ *
+ * - no additional memory required
+ */
+static int esdm_rpcs_pack_internal(const ProtobufCMessage *message,
+				   struct esdm_rpcs_connection *rpc_conn)
+{
+	struct esdm_rpc_proto_sc_header sc_header;
+	struct esdm_rpcs_write_buf tmp = { 0 };
+	size_t message_length;
+	int ret;
+
+	message_length = protobuf_c_message_get_packed_size(message);
+	tmp.base.append = esdm_rpcs_append_data;
+	tmp.rpc_conn = rpc_conn;
+
+	sc_header.status_code = le_bswap32(PROTOBUF_C_RPC_STATUS_CODE_SUCCESS);
+	sc_header.method_index = le_bswap32(rpc_conn->method_index);
+	sc_header.message_length = le_bswap32(message_length);
+	sc_header.request_id = le_bswap32(rpc_conn->request_id);
+
+	esdm_logger(
+		LOGGER_DEBUG, LOGGER_C_RPC,
+		"Server sending: server status %u, message length %u, message index %u, request ID %u\n",
+		sc_header.status_code, sc_header.message_length,
+		sc_header.method_index, sc_header.request_id);
+
+	CKINT_LOG(esdm_rpcs_write_data(rpc_conn, (uint8_t *)&sc_header,
+				       sizeof(sc_header)),
+		  "Submission of header data failed with error %d\n", ret);
+
+	if (protobuf_c_message_pack_to_buffer(message, &tmp.base) !=
+	    message_length) {
+		esdm_logger(LOGGER_VERBOSE, LOGGER_C_RPC,
+			    "Short write of data to file descriptor\n");
+		ret = -EFAULT;
+	}
+
+out:
+	return ret;
+}
+
+#endif /* ESDM_RPCS_BUF_WRITE */
+
 /* Pack the message into a ProtobufC structure and write it to the receiver. */
 static int esdm_rpcs_pack(const ProtobufCMessage *message,
 			  struct esdm_rpcs_connection *rpc_conn)
 {
 	struct esdm_rpc_proto_sc_header sc_header;
-	struct esdm_rpcs_write_buf tmp = { 0 };
-	size_t message_length;
-	int ret;
 
 	if (!protobuf_c_message_check(message)) {
 		sc_header.status_code =
@@ -177,33 +307,7 @@ static int esdm_rpcs_pack(const ProtobufCMessage *message,
 					    sizeof(sc_header));
 	}
 
-	message_length = protobuf_c_message_get_packed_size(message);
-	tmp.base.append = esdm_rpcs_append_data;
-	tmp.rpc_conn = rpc_conn;
-
-	sc_header.status_code = le_bswap32(PROTOBUF_C_RPC_STATUS_CODE_SUCCESS);
-	sc_header.method_index = le_bswap32(rpc_conn->method_index);
-	sc_header.message_length = le_bswap32(message_length);
-	sc_header.request_id = le_bswap32(rpc_conn->request_id);
-
-	logger(LOGGER_DEBUG, LOGGER_C_RPC,
-	       "Server sending: server status %u, message length %u, message index %u, request ID %u\n",
-	       sc_header.status_code, sc_header.message_length,
-	       sc_header.method_index, sc_header.request_id);
-
-	CKINT_LOG(esdm_rpcs_write_data(rpc_conn, (uint8_t *)&sc_header,
-				       sizeof(sc_header)),
-		  "Submission of header data failed with error %d\n", ret);
-
-	if (protobuf_c_message_pack_to_buffer(message, &tmp.base) !=
-	    message_length) {
-		logger(LOGGER_VERBOSE, LOGGER_C_RPC,
-		       "Short write of data to file descriptor \n");
-		ret = -EFAULT;
-	}
-
-out:
-	return ret;
+	return esdm_rpcs_pack_internal(message, rpc_conn);
 }
 
 /* Is the calling RPC client a privileged user? */
@@ -218,12 +322,13 @@ bool esdm_rpc_client_is_privileged(void *closure_data)
 		return false;
 
 	if (cred.uid == 0) {
-		logger(LOGGER_DEBUG, LOGGER_C_ANY,
-		       "Remote client is privileged\n");
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_ANY,
+			    "Remote client is privileged\n");
 		return true;
 	}
 
-	logger(LOGGER_DEBUG, LOGGER_C_ANY, "Remote client is not privileged\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_ANY,
+		    "Remote client is not privileged\n");
 	return false;
 }
 
@@ -271,6 +376,10 @@ out:
 		protobuf_c_message_free_unpacked(message,
 						 rpc_conn->rpc_allocator);
 
+	/* Pick up the error from esdm_rpcs_write_data */
+	if (rpc_conn->child_fd == -1)
+		ret = -EPIPE;
+
 	return ret;
 }
 
@@ -327,9 +436,9 @@ static int esdm_rpcs_read(struct esdm_rpcs_connection *rpc_conn)
 		total_received += (size_t)received;
 		buf_p += (size_t)received;
 
-		logger(LOGGER_DEBUG, LOGGER_C_ANY,
-		       "Reading %zd bytes, already consumed %zu bytes\n",
-		       received, total_received);
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_ANY,
+			    "Reading %zd bytes, already consumed %zu bytes\n",
+			    received, total_received);
 
 		/* We insist on having at least a header received. */
 		if (total_received < sizeof(*received_data))
@@ -346,10 +455,11 @@ static int esdm_rpcs_read(struct esdm_rpcs_connection *rpc_conn)
 			header->method_index = le_bswap32(header->method_index);
 			header->request_id = le_bswap32(header->request_id);
 
-			logger(LOGGER_DEBUG, LOGGER_C_RPC,
-			       "Server received: message length %u, message index %u, request ID %u\n",
-			       header->message_length, header->method_index,
-			       header->request_id);
+			esdm_logger(
+				LOGGER_DEBUG, LOGGER_C_RPC,
+				"Server received: message length %u, message index %u, request ID %u\n",
+				header->message_length, header->method_index,
+				header->request_id);
 
 			/*
 			 * Truncate the buffer length if client specified
@@ -423,8 +533,9 @@ static int esdm_rpcs_handler(void *args)
 		ret = esdm_rpcs_read(rpc_conn);
 	} while (!ret);
 
-	logger(LOGGER_DEBUG, LOGGER_C_RPC,
-	       "Closing incoming connection for FD %d\n", rpc_conn->child_fd);
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_RPC,
+		    "Closing incoming connection for FD %d\n",
+		    rpc_conn->child_fd);
 	esdm_rpcs_release_conn(rpc_conn);
 	return 0;
 }
@@ -449,8 +560,9 @@ static int esdm_rpcs_workerloop(struct esdm_rpcs *proto)
 	int ret = 0;
 
 #ifdef DEBUG
-	logger(LOGGER_WARN, LOGGER_C_RPC,
-	       "Debug mode enabled, RPC server executes single threaded\n");
+	esdm_logger(
+		LOGGER_WARN, LOGGER_C_RPC,
+		"Debug mode enabled, RPC server executes single threaded\n");
 #endif
 
 	if (proto->server_listening_fd < 0)
@@ -476,8 +588,57 @@ static int esdm_rpcs_workerloop(struct esdm_rpcs *proto)
 			continue;
 		}
 
+		/* If server is requested to terminate, do that */
+		if (atomic_read(&server_exit))
+			break;
+
 		rpc_conn->proto = proto;
 
+		/*
+		 * When there is a high load, there is a race between SIGTERM
+		 * signal breaking the accept() call and the propagation that
+		 * server_exit is set to true. Thus, in a shutdown under high
+		 * load, it is possible that the caller triggered yet another
+		 * accept() before server_exit is set. In this case, accept
+		 * now blocks indefinitely. If it is unacceptable to check
+		 * accept's errno for EINTR and then terminate the loop,
+		 * because other signals than SIGTERM can be received that
+		 * should not immediately terminate the worker loop, undefine
+		 * this macro to use a timer-based wakeup to check for the
+		 * status of server_exit.
+		 */
+#ifndef ESDM_WORKERLOOP_TERM_ON_SIGNAL
+		for (;;) {
+			struct timeval timeout = { .tv_sec = 1, .tv_usec = 0 };
+			fd_set fds;
+			int sret;
+
+			FD_ZERO(&fds);
+			FD_SET(proto->server_listening_fd, &fds);
+
+			sret = select(proto->server_listening_fd + 1, &fds,
+				      NULL, NULL, &timeout);
+
+			/* error */
+			if (sret == -1 && errno != EINTR) {
+				esdm_logger(LOGGER_ERR, LOGGER_C_ANY,
+					    "Select returned with error %s\n",
+					    strerror(errno));
+				goto out;
+			}
+
+			/* activity */
+			if (sret > 0)
+				break;
+
+			/* timeout - yet server shall exit */
+			if (atomic_read(&server_exit))
+				goto out;
+
+			/* timeout - simply retry */
+		}
+#endif
+
 		/* Wait for incoming connection */
 		rpc_conn->child_fd =
 			accept(proto->server_listening_fd, &addr, &addr_len);
@@ -487,11 +648,18 @@ static int esdm_rpcs_workerloop(struct esdm_rpcs *proto)
 			break;
 
 		if (rpc_conn->child_fd < 0) {
+#ifdef ESDM_WORKERLOOP_TERM_ON_SIGNAL
+			/* Terminate the worker loop upon receipt of signal */
+			if (errno == EINTR)
+				goto out;
+#endif
+
 			esdm_rpcs_release_conn(rpc_conn);
 			rpc_conn = NULL;
-			logger(LOGGER_WARN, LOGGER_C_ANY,
-			       "Accepting incoming connections failed: %s\n",
-			       strerror(errno));
+			esdm_logger(
+				LOGGER_WARN, LOGGER_C_ANY,
+				"Accepting incoming connections failed: %s\n",
+				strerror(errno));
 			continue;
 		}
 
@@ -501,17 +669,17 @@ static int esdm_rpcs_workerloop(struct esdm_rpcs *proto)
 			       (const char *)&tv, sizeof(tv)) < 0) {
 			int errsv = errno;
 
-			logger(LOGGER_ERR, LOGGER_C_RPC,
-			       "Error setting timeout on socket: %s\n",
-			       strerror(errsv));
+			esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+				    "Error setting timeout on socket: %s\n",
+				    strerror(errsv));
 			esdm_rpcs_release_conn(rpc_conn);
 			rpc_conn = NULL;
 			continue;
 		}
 
-		logger(LOGGER_DEBUG, LOGGER_C_RPC,
-		       "Processing new incoming connection for FD %d\n",
-		       rpc_conn->child_fd);
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_RPC,
+			    "Processing new incoming connection for FD %d\n",
+			    rpc_conn->child_fd);
 
 		/* Handle new incoming connection */
 #ifdef DEBUG
@@ -522,8 +690,9 @@ static int esdm_rpcs_workerloop(struct esdm_rpcs *proto)
 		esdm_rpcs_handler(rpc_conn);
 #else /* DEBUG */
 		if (thread_start(esdm_rpcs_handler, rpc_conn, 0, NULL)) {
-			logger(LOGGER_ERR, LOGGER_C_RPC,
-			       "Starting new thread for incoming connection failed\n");
+			esdm_logger(
+				LOGGER_ERR, LOGGER_C_RPC,
+				"Starting new thread for incoming connection failed\n");
 			esdm_rpcs_release_conn(rpc_conn);
 			rpc_conn = NULL;
 			continue;
@@ -568,29 +737,35 @@ static int esdm_rpcs_start(const char *unix_socket, uint16_t tcp_port,
 		return -EINVAL;
 	}
 
-	fd = socket(protocol_family, SOCK_SEQPACKET, 0);
+	fd = socket(protocol_family,
+		    SOCK_SEQPACKET
+#ifndef ESDM_WORKERLOOP_TERM_ON_SIGNAL
+			    | SOCK_NONBLOCK | SOCK_CLOEXEC
+#endif
+		    ,
+		    0);
 	if (fd < 0) {
 		errsv = -errno;
-		logger(LOGGER_ERR, LOGGER_C_RPC,
-		       "RPC Server: cannot create socket: %s\n",
-		       strerror(errsv));
+		esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+			    "RPC Server: cannot create socket: %s\n",
+			    strerror(errsv));
 		return -errsv;
 	}
 
 	if (bind(fd, address, address_len) < 0) {
 		errsv = -errno;
-		logger(LOGGER_ERR, LOGGER_C_RPC,
-		       "RPC Server: cannot bind to socket: %s\n",
-		       strerror(errsv));
+		esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+			    "RPC Server: cannot bind to socket: %s\n",
+			    strerror(errsv));
 		close(fd);
 		return -errsv;
 	}
 
 	if (listen(fd, 255) < 0) {
 		errsv = -errno;
-		logger(LOGGER_ERR, LOGGER_C_RPC,
-		       "RPC Server: cannot listen on socket: %s\n",
-		       strerror(errsv));
+		esdm_logger(LOGGER_ERR, LOGGER_C_RPC,
+			    "RPC Server: cannot listen on socket: %s\n",
+			    strerror(errsv));
 		close(fd);
 		return -errsv;
 	}
@@ -635,9 +810,10 @@ static int esdm_rpcs_unpriv_init(void *args)
 	    -1) {
 		ret = -errno;
 
-		logger(LOGGER_ERR, LOGGER_C_ANY,
-		       "Failed to set permissions for Unix domain socket %s: %s\n",
-		       ESDM_RPC_UNPRIV_SOCKET, strerror(errno));
+		esdm_logger(
+			LOGGER_ERR, LOGGER_C_ANY,
+			"Failed to set permissions for Unix domain socket %s: %s\n",
+			ESDM_RPC_UNPRIV_SOCKET, strerror(errno));
 
 		goto out;
 	}
@@ -650,9 +826,9 @@ static int esdm_rpcs_unpriv_init(void *args)
 	thread_wait_event(&esdm_rpc_thread_init_wait,
 			  (atomic_read(&esdm_rpc_init_state) ==
 			   esdm_rpcs_state_perm_dropped));
-	logger(LOGGER_DEBUG, LOGGER_C_RPC,
-	       "Unprivileged server thread for %s available\n",
-	       ESDM_RPC_UNPRIV_SOCKET);
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_RPC,
+		    "Unprivileged server thread for %s available\n",
+		    ESDM_RPC_UNPRIV_SOCKET);
 
 	/* Server handing unprivileged interface in current thread */
 	CKINT(esdm_rpcs_workerloop(&unpriv_proto));
@@ -690,9 +866,10 @@ static int esdm_rpcs_interfaces_init(const char *username)
 	if (chmod(ESDM_RPC_PRIV_SOCKET, S_IRUSR | S_IWUSR) == -1) {
 		int errsv = errno;
 
-		logger(LOGGER_ERR, LOGGER_C_ANY,
-		       "Failed to set permissions for Unix domain socket %s: %s\n",
-		       ESDM_RPC_PRIV_SOCKET, strerror(errsv));
+		esdm_logger(
+			LOGGER_ERR, LOGGER_C_ANY,
+			"Failed to set permissions for Unix domain socket %s: %s\n",
+			ESDM_RPC_PRIV_SOCKET, strerror(errsv));
 		ret = -errsv;
 		goto out;
 	}
@@ -713,9 +890,9 @@ static int esdm_rpcs_interfaces_init(const char *username)
 	/* Notify all unpriv handler threads that they can become active */
 	atomic_set(&esdm_rpc_init_state, esdm_rpcs_state_perm_dropped);
 	thread_wake_all(&esdm_rpc_thread_init_wait);
-	logger(LOGGER_DEBUG, LOGGER_C_RPC,
-	       "Privileged server thread for %s available\n",
-	       ESDM_RPC_PRIV_SOCKET);
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_RPC,
+		    "Privileged server thread for %s available\n",
+		    ESDM_RPC_PRIV_SOCKET);
 
 	/* Server handing privileged interface in current thread */
 	CKINT(esdm_rpcs_workerloop(&priv_proto));
@@ -732,24 +909,26 @@ static void esdm_rpcs_cleanup(void)
 {
 	/* Clean up all unprivileged Unix domain socket */
 	if (unlink(ESDM_RPC_UNPRIV_SOCKET) < 0) {
-		logger(LOGGER_ERR, LOGGER_C_SERVER,
-		       "ESDM Unix domain socket %s cannot be deleted: %s\n",
-		       ESDM_RPC_UNPRIV_SOCKET, strerror(errno));
+		esdm_logger(
+			LOGGER_ERR, LOGGER_C_SERVER,
+			"ESDM Unix domain socket %s cannot be deleted: %s\n",
+			ESDM_RPC_UNPRIV_SOCKET, strerror(errno));
 	} else {
-		logger(LOGGER_DEBUG, LOGGER_C_SERVER,
-		       "ESDM Unix domain socket %s deleted\n",
-		       ESDM_RPC_UNPRIV_SOCKET);
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_SERVER,
+			    "ESDM Unix domain socket %s deleted\n",
+			    ESDM_RPC_UNPRIV_SOCKET);
 	}
 
 	/* Clean up the privileged Unix domain socket */
 	if (unlink(ESDM_RPC_PRIV_SOCKET) < 0) {
-		logger(LOGGER_ERR, LOGGER_C_SERVER,
-		       "ESDM Unix domain socket %s cannot be deleted: %s\n",
-		       ESDM_RPC_PRIV_SOCKET, strerror(errno));
+		esdm_logger(
+			LOGGER_ERR, LOGGER_C_SERVER,
+			"ESDM Unix domain socket %s cannot be deleted: %s\n",
+			ESDM_RPC_PRIV_SOCKET, strerror(errno));
 	} else {
-		logger(LOGGER_DEBUG, LOGGER_C_SERVER,
-		       "ESDM Unix domain socket %s deleted\n",
-		       ESDM_RPC_PRIV_SOCKET);
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_SERVER,
+			    "ESDM Unix domain socket %s deleted\n",
+			    ESDM_RPC_PRIV_SOCKET);
 	}
 
 	/*
@@ -766,26 +945,26 @@ static void esdm_rpcs_cleanup(void)
 	esdm_shmid = shmget(key, sizeof(struct esdm_shm_status),
 			    S_IRUSR | S_IRGRP | S_IROTH);
 	if (esdm_shmid < 0) {
-		logger(LOGGER_ERR, LOGGER_C_SERVER,
+		esdm_logger(LOGGER_ERR, LOGGER_C_SERVER,
 		       "ESDM shared memory segment attachment for deletion failed: %s\n",
 		       strerror(errno));
 	} else {
 		if (shmctl(esdm_shmid, IPC_RMID, NULL) < 0) {
-			logger(LOGGER_ERR, LOGGER_C_SERVER,
+			esdm_logger(LOGGER_ERR, LOGGER_C_SERVER,
 			       "ESDM shared memory segment cannot be deleted: %s\n",
 			       strerror(errno));
 		} else {
-			logger(LOGGER_DEBUG, LOGGER_C_SERVER,
+			esdm_logger(LOGGER_DEBUG, LOGGER_C_SERVER,
 			       "ESDM shared memory segment deleted\n");
 		}
 	}
 
 	/* Clean up the status semaphore */
 	if (sem_unlink(ESDM_SEM_NAME)) {
-		logger(LOGGER_VERBOSE, LOGGER_C_SERVER,
+		esdm_logger(LOGGER_VERBOSE, LOGGER_C_SERVER,
 		       "Cannot unlink semaphore: %s\n", strerror(errno));
 	} else {
-		logger(LOGGER_DEBUG, LOGGER_C_SERVER,
+		esdm_logger(LOGGER_DEBUG, LOGGER_C_SERVER,
 		       "ESDM semaphore deleted\n");
 	}
 #endif
@@ -814,8 +993,8 @@ static void esdm_rpc_priv_init_complete(void)
 	if (atomic_read(&esdm_rpc_init_state) != esdm_rpcs_state_uninitialized)
 		return;
 
-	logger(LOGGER_DEBUG, LOGGER_C_SERVER,
-	       "Privileged intiialization complete\n");
+	esdm_logger(LOGGER_DEBUG, LOGGER_C_SERVER,
+		    "Privileged intiialization complete\n");
 	/* Notification that the privileged initialization is complete. */
 	atomic_set(&esdm_rpc_init_state, esdm_rpcs_state_priv_init_complete);
 	thread_wake_all(&esdm_rpc_thread_init_wait);
@@ -844,8 +1023,8 @@ int esdm_rpc_server_init(const char *username)
 
 	pid = fork();
 	if (pid < 0) {
-		logger(LOGGER_ERR, LOGGER_C_SERVER,
-		       "Cannot fork interface process\n");
+		esdm_logger(LOGGER_ERR, LOGGER_C_SERVER,
+			    "Cannot fork interface process\n");
 		exit(1);
 	} else if (pid == 0) {
 		pthread_setname_np(pthread_self(), "ESDM master");
@@ -853,8 +1032,8 @@ int esdm_rpc_server_init(const char *username)
 		/* Create thread for entropy source monitor */
 		if (thread_start(esdm_rpc_server_es_monitor, NULL,
 				 ESDM_THREAD_ES_MONITOR, NULL)) {
-			logger(LOGGER_WARN, LOGGER_C_RPC,
-			       "Starting ES monitor thread failed\n");
+			esdm_logger(LOGGER_WARN, LOGGER_C_RPC,
+				    "Starting ES monitor thread failed\n");
 		}
 
 		/* Wait for the privileged initialization to complete. */
@@ -903,6 +1082,8 @@ out:
 
 void esdm_rpc_server_fini(void)
 {
+	pid_t tmp;
+
 	thread_stop_spawning();
 
 	atomic_set(&server_exit, 1);
@@ -910,6 +1091,14 @@ void esdm_rpc_server_fini(void)
 
 	/* Unblock the accept() in the server loop */
 	thread_send_signal(ESDM_THREAD_RPC_UNPRIV_GROUP, SIGUSR1);
+	thread_send_signal(ESDM_THREAD_CUSE_POLL_GROUP, SIGUSR1);
+	/* Unblock the accept() in the privileged server loop.
+	 * Only some signal handled by the signal handler before fork
+	 * will be handled in the forked child and will set server_exit there.
+	 * Setting server_exit in the main process will have no effect */
+	tmp = server_pid;
+	if (tmp > 1)
+		kill(tmp, SIGTERM);
 
 	/* Terminate test pertubation support */
 	esdm_test_shm_status_fini();
